(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_graph.ml,v 1.13 2014-02-10 18:12:49 fclement Exp $ *)

(* Matrix graphs of function. *)

module Nat = Unatural;;

module Idx = Uindex_index;;
module Set = Umatrix_nonempty_set;;

let mg_qual = Upervasives.qualify "Ugraph";;

let mg_invalid_arg = Upervasives.invalid_arg_qual mg_qual;;
let mg_failwith = Upervasives.failwith_qual mg_qual;;

type index = Set.index
and uindex = Set.uindex
;;

type domain = Set.t;;

type sub_domain = Set.nonempty_subset;;

type 'a images =
  | Constant of 'a
  | Images of 'a array
;;

type 'a t =
  {
   domain : domain;
   images : 'a images;
  }
;;

let same = ( == );;

let same_domain mg1 mg2 = Set.same mg1.domain mg2.domain;;

let equal_domain mg1 mg2 = Set.equal mg1.domain mg2.domain;;

let unique_domain mg1 mg2 =
  if not (equal_domain mg1 mg2) then
    mg_failwith "unique_domain" "domains differ" else
    { mg2 with domain = mg1.domain; }
;;

(* Construction functions. *)

let array_sub_for_all p a ofs len =
  let rec loop j i = j = len || p a.(i) && loop (Nat.succ j) (succ i) in
  loop Nat._0_ ofs
;;

let detect_constant_images mg =
  match mg.images with
  | Constant _ -> mg
  | Images ys ->
      let len = Uarray.length ys in
      if len = Nat._0_ then assert false else
      let y0 = ys.(0) in
      let p y = y = y0 in
      let b = array_sub_for_all p ys 1 (Nat.pred len) in
      if b then { mg with images = Constant y0; } else mg
;;

let make m =
  let sz = Umatrix.size m in
  let dom = Set.make sz in
  let ys = Umatrix.to_array m in
  detect_constant_images { domain = dom; images = Images ys; }
;;

let constant dom y = { domain = dom; images = Constant y; };;

let unsafe_from_array dom ys =
  detect_constant_images { domain = dom; images = Images ys; }
;;

let from_array dom ys =
  if Set.cardinal dom <> Uarray.length ys then
    mg_invalid_arg "from_array" "arguments have different sizes" else
  unsafe_from_array dom ys
;;

let from_fun dom f =
  let card = Set.cardinal dom in
  let (idx0, uidx0) = Set.choice dom in
  let y0 = f idx0 uidx0 in
  let ys = Uarray.make card y0 in
  let ff idx uidx = ys.(Idx.to_int idx) <- f idx uidx in
  Set.sub_iter ff dom Idx._1_ (Nat.pred card);
  unsafe_from_array dom ys
;;

(* Access functions. *)

let aux_constant fun_name mg f =
  match mg.images with
  | Constant y -> f mg y
  | Images _ -> mg_invalid_arg fun_name "non constant graph"
;;

let get_constant mg = aux_constant "get_constant" mg (fun _ y -> y);;

let change_constant mg y =
  aux_constant "change_constant"
    mg (fun mg _ -> { mg with images = Constant y; })
;;

let image mg =
  match mg.images with
  | Constant y -> (fun _idx -> y)
  | Images ys -> (fun idx -> ys.(Idx.to_int idx))
;;

let to_array = function
  | { domain = dom; images = Constant y; } -> Uarray.make (Set.cardinal dom) y
  | { images = Images a; _; } -> a
;;

let restriction
    ({ domain = dom; images = images; }  as mg)
    { Set.support = sup; indices = indices; nonempty_set = sub_dom; } =
  if sup != dom then
    mg_invalid_arg "restriction" "sub-domain is not a subset of domain" else
  match images with
  | Constant y -> constant sub_dom y
  | Images _ -> from_array sub_dom (Array.map (image mg) indices)
;;

(* Iterators. *)

let map_images f mg =
  match mg.images with
  | Constant y -> { mg with images = Constant (f y); }
  | Images ys -> { mg with images = Images (Array.map f ys); }
;;
